/*
 * RenderArea.cpp
 * 
 * Copyright (C) 2008 Matthias Schiffer <matthias@gamezock.de>
 * 
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "RenderArea.h"
#include <View/View.h>
#include <iostream>
#include <cstdlib>
#include <cmath>
#include <gtkmm/toolbutton.h>
#include <gtkmm/adjustment.h>
#include <GL/gl.h>

namespace ZoomEdit {
namespace Gui {

GdkGLConfig *RenderArea::glconfig = 0;

RenderArea::RenderArea(BaseObjectType *cobject, const Glib::RefPtr<Gnome::Glade::Xml> &xml)
: Gtk::DrawingArea(cobject), view(0), xCenter(0), yCenter(0), viewWidth(0), viewHeight(0), zoomLevel(0),
scale(100), inWindow(false) {
  if(!glconfig) {
    glconfig = gdk_gl_config_new_by_mode((GdkGLConfigMode)(GDK_GL_MODE_RGB | GDK_GL_MODE_DOUBLE));
    if(!glconfig) glconfig = gdk_gl_config_new_by_mode(GDK_GL_MODE_RGB); // Hmm, can't find double buffered config
    if(!glconfig) {
      std::cerr << "*** No appropriate OpenGL-capable visual found." << std::endl;
      std::exit(1);
    }
  }
  
  signal_realize().connect(sigc::mem_fun(this, &RenderArea::onRealize));
  signal_configure_event().connect(sigc::mem_fun(this, &RenderArea::onConfigureEvent));
  signal_expose_event().connect(sigc::mem_fun(this, &RenderArea::onExposeEvent));
  signal_scroll_event().connect(sigc::mem_fun(this, &RenderArea::onScrollEvent));
  signal_enter_notify_event().connect(sigc::mem_fun(this, &RenderArea::onEnterNotifyEvent));
  signal_leave_notify_event().connect(sigc::mem_fun(this, &RenderArea::onLeaveNotifyEvent));
  signal_motion_notify_event().connect(sigc::mem_fun(this, &RenderArea::onMotionNotifyEvent));
  
  Gtk::ToolButton *button;
  xml->get_widget("ToolButtonZoomIn", button);
  button->signal_clicked().connect(sigc::bind(sigc::mem_fun(this, &RenderArea::zoom), 2, 0.5f, 0.5f));
  
  xml->get_widget("ToolButtonZoomOut", button);
  button->signal_clicked().connect(sigc::bind(sigc::mem_fun(this, &RenderArea::zoom), -2, 0.5f, 0.5f));
  
  gtk_widget_set_gl_capability(GTK_WIDGET(cobject), glconfig, 0, TRUE, GDK_GL_RGBA_TYPE);
}

void RenderArea::onRealize() {
  if(!gdkGLBegin())
    return;
  
  glClearColor(0, 0, 0, 0);
  
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  
  glEnable(GL_LINE_SMOOTH);
  glEnable(GL_POINT_SMOOTH);
  
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  gdkGLEnd();
}

bool RenderArea::onConfigureEvent(GdkEventConfigure*) {
  if(!gdkGLBegin())
    return false;
  
  glViewport(0, 0, get_width(), get_height());
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  
  if(get_width() != 0 && get_height() != 0)
    glScalef(2.0f/get_width(), -2.0f/get_height(), 1);
  
  gdkGLEnd();
  
  viewWidth = get_width()/scale;
  viewHeight = get_height()/scale;
  
  queue_draw();
  
  return true;
}

bool RenderArea::onExposeEvent(GdkEventExpose*) {
  if(!gdkGLBegin())
    return false;
  
  glClear(GL_COLOR_BUFFER_BIT);
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  
  glScalef(scale, scale, 1);
  glTranslatef(-xCenter, -yCenter, 0);
  
  if(view)
    view->render(this);
  
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  
  gdkSwapBuffers();
  gdkGLEnd();
  
  return true;
}

bool RenderArea::onScrollEvent(GdkEventScroll *event) {
  switch(event->direction) {
    case GDK_SCROLL_UP:
      zoom(1, event->x/get_width(), event->y/get_height());
      return true;
    case GDK_SCROLL_DOWN:
      zoom(-1, event->x/get_width(), event->y/get_height());
      return true;
    default:
      return false;
  }
}

bool RenderArea::onEnterNotifyEvent(GdkEventCrossing *event) {
  inWindow = true;
  xHover = event->x/scale - getViewWidth()/2 + xCenter;
  yHover = event->y/scale - getViewHeight()/2 + yCenter;
  
  return true;
}

bool RenderArea::onLeaveNotifyEvent(GdkEventCrossing*) {
  inWindow = false;
  
  return true;
}

bool RenderArea::onMotionNotifyEvent(GdkEventMotion *event) {
  inWindow = true;
  
  if(event->state & GDK_BUTTON3_MASK) {
    xCenter = xHover - event->x/scale + getViewWidth()/2;
    yCenter = yHover - event->y/scale + getViewHeight()/2;
    queue_draw();
  }
  else {
    xHover = event->x/scale - getViewWidth()/2 + xCenter;
    yHover = event->y/scale - getViewHeight()/2 + yCenter;
  }
  
  return true;
}

void RenderArea::zoom(int zoom, float x, float y) {
  zoomLevel = std::max(std::min(zoomLevel + zoom, 50), -100);
  scale = 100*std::pow(1.1f, zoomLevel);
  
  const float newWidth = get_width()/scale, newHeight = get_height()/scale;
  
  xCenter += (x-0.5f)*(viewWidth-newWidth);
  yCenter += (y-0.5f)*(viewHeight-newHeight);
  
  viewWidth = newWidth;
  viewHeight = newHeight;
  
  queue_draw();
}

}
}
